from albatross.http_error import HTTPError
from albatross.status_codes import HTTP_400


def caseless_pairs(seq):
    for k, v in seq:
        yield k.lower(), v


class Immutable:
    def __setitem__(self, k, v):
        raise TypeError('Cannot set item on %s' % self.__class__)

    def update(self, E=None, **F):
        raise TypeError('Cannot update on %s' % self.__class__)


class ImmutableMultiDict(Immutable, dict):
    def __getitem__(self, k):
        if k in self:
            return super(ImmutableMultiDict, self).__getitem__(k)[0]
        raise HTTPError(HTTP_400, 'Must provide parameter \'%s\'.' % k)

    def get(self, k, d=None):
        if k in self:
            return super(ImmutableMultiDict, self).__getitem__(k)[0]
        return d

    def get_all(self, k, d=None):
        if k in self:
            return super(ImmutableMultiDict, self).__getitem__(k)
        return d


class CaselessDict(dict):
    def __init__(self, *args, **kwargs):
        if args:
            args = caseless_pairs(args)
        if kwargs:
            kwargs = {k.lower(): v for k, v in kwargs.items()}
        super(CaselessDict, self).__init__(args, **kwargs)

    def __contains__(self, k):
        return super(CaselessDict, self).__contains__(k.lower())

    def __getitem__(self, k):
        return super(CaselessDict, self).__getitem__(k.lower())

    def __iter__(self):
        for k in super(CaselessDict, self).__iter__():
            yield k.lower()

    def __setitem__(self, k, v):
        super(CaselessDict, self).__setitem__(k.lower(), v)

    def get(self, k, d=None):
        if k in self:
            return super(CaselessDict, self).__getitem__(k.lower())
        return d

    def update(self, other=None, **kwargs):
        updates = {k.lower(): v for k, v in kwargs.items()}
        if other:
            if hasattr(other, 'items'):
                other = other.items()
            updates.update(caseless_pairs(other))
        return super(CaselessDict, self).update(updates)


class ImmutableCaselessDict(Immutable, CaselessDict):
    pass


class ImmutableCaselessMultiDict(ImmutableMultiDict, CaselessDict):
    pass



from albatross.data_types import ImmutableMultiDict
from albatross.compat import json
import urllib.parse as parse
import cgi
import io


def trim_keys(d):
    return {k.strip(): v for k, v in d.items()}


class FileStorage:

    def __init__(self, field_storage):
        self.filename = field_storage.filename
        self.value = field_storage.value


class Request:
    """
    Attributes:
        method (str): One of GET, POST, PUT, DELETE
        path (str): Full request path
        query_string (str): Full query string
        query (dict): dict of the request query
        body (str): Request body
        args (dict): Dictionary of named parameters in route regex
        form (dict): Dictionary of body parameters
    """

    def __init__(self, method, path, query_string, raw_body, args, headers):
        self.method = method
        self.path = path
        self.query_string = query_string
        self.query = ImmutableMultiDict(parse.parse_qs(query_string))
        self.args = args
        self.headers = headers
        self.raw_body = None
        self.form = None
        if raw_body:
            self._parse_body(raw_body)
        if 'Cookie' in headers:
            self.cookies = self._parse_cookie(headers['Cookie'])
        else:
            self.cookies = ImmutableMultiDict()

    def _parse_cookie(self, value):
        cookies = trim_keys(parse.parse_qs(value))
        return ImmutableMultiDict(cookies)

    def _parse_form(self, raw_body):
        # TODO theres probably a way to not read whole body first)
        fp = io.BytesIO(raw_body)
        form = cgi.FieldStorage(fp, headers=self.headers, environ={'REQUEST_METHOD': 'POST'})
        d = {}
        for k in form.keys():
            if form[k].filename:
                d[k] = [FileStorage(form[k])]
            else:
                d[k] = [form[k].value]
        return ImmutableMultiDict(d)

    def _parse_body(self, raw_body):
        content_type = self.headers.get('Content-Type', '')
        if content_type == 'application/json':
            self.form = json.loads(raw_body.decode())
        elif content_type.startswith('multipart/form-data'):
            self.form = self._parse_form(raw_body)
        elif content_type == 'application/x-www-form-urlencoded':
            self.form = ImmutableMultiDict(parse.parse_qs(raw_body.decode()))
        else:
            self.raw_body = raw_body
